/*
 *  debugafs - file system debugger for AtheOS
 * 
 *  Copyright (C) 2001 Mike Bennett (mike@stressbunny.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
*/

#include "debugafs.h"
#include "afs.h"
#include "stat.h"	// Kludge, remove once compiling on AtheOS

#include<stdio.h>
#include<stdlib.h>
#include<string.h>

extern int open(const char *, int);
extern int close(int);

void cmd_show_help(char **paCmdtokens, int nWhich)
{
	switch(nWhich) {
		case HELP_UNKNOWN:
			printf("Unknown command, use ? to get help\n");
		break;
		
		case HELP_OPEN:
			printf("Usage: open <device>\n");
		break;

		case HELP_SHOW_INODE_STATS:
			printf("Usage: inode_stats <block> <group>\n");
		break;
		
		default:
			printf("\ndebugafs commands\n");
			printf("-----------------\n");
			printf("?\t\t\t- Show the help\n");
			printf("v\t\t\t- Show the version\n");
			printf("open <device>\t\t- Open a device\n");
			printf("close\t\t\t- Close the opened device\n");
			printf("status\t\t\t- Display current settings, including open devices, etc\n");
			printf("super_stats [pos]\t- Show the superblock stats, pos = position to read\n");
			printf("inode_stats <block> <group>\t- Show inode stats\n");
			printf("\nq\t\t\t- Quit\n\n");
	}
}

void cmd_show_version(char **paCmdtokens, void *dummy)
{
	printf("\nAtheOS AFS debugger - version %s\n", VERSION);
	printf("===================================\n");
	printf("Mike Bennett (mike@stressbunny.com)\n\n");
	printf("- http://www.stressbunny.com/atheos\n\n");
}

void cmd_show_status(char **paCmdtokens, DebugAFS_s *psDebugAFS)
{
	printf("\nStatus\n------\n");
	printf("Open device: %s\n", psDebugAFS->de_nDev != DFS_DEVICE_CLOSED ? psDebugAFS->de_pzDevice : "NONE");
	printf("Super block read: %s\n", psDebugAFS->de_psVolume->av_nSuperBlockPos != DFS_NOREAD_SBLOCK ? "YES" : "NO");
	printf("\n");
}

void cmd_open_device(char **paCmdtokens, DebugAFS_s *psDebugAFS)
{
	if(psDebugAFS->de_nDev != DFS_DEVICE_CLOSED) {
		printf("Error: Device %s is already open, please close that first\n", psDebugAFS->de_pzDevice);
		return;
	}

	if(num_tokens(paCmdtokens) != 2) {
		cmd_show_help(NULL, HELP_OPEN);
		return;
	}

	psDebugAFS->de_pzDevice = strdup(paCmdtokens[1]);
	psDebugAFS->de_nDev = open(psDebugAFS->de_pzDevice, O_RDONLY);

	if(psDebugAFS->de_nDev < 0) {
		printf("Error: Cannot open the device: %s\n", psDebugAFS->de_pzDevice);
		return;
	}

	printf("Opened the device: %s\n", psDebugAFS->de_pzDevice);
}

void cmd_close_device(char **paCmdtokens, DebugAFS_s *psDebugAFS)
{
	if(psDebugAFS->de_nDev == DFS_DEVICE_CLOSED) {
		printf("Error: No device currently open\n");
		return;
	}

	if(close(psDebugAFS->de_nDev) > 0) {
		printf("Error: Cannot close the device: %s\n", psDebugAFS->de_pzDevice);
		return;
	}

	printf("Closed device: %s\n", psDebugAFS->de_pzDevice);
	psDebugAFS->de_nDev = DFS_DEVICE_CLOSED;

	psDebugAFS->de_psVolume->av_nSuperBlockPos = DFS_NOREAD_SBLOCK;
}

void cmd_show_superblock_stats(char **paCmdtokens, DebugAFS_s *psDebugAFS)
{
	if(psDebugAFS->de_nDev == DFS_DEVICE_CLOSED) {
		printf("Error: No device currently open, please use open\n");
		return;
	}

	/* Are we being told a position to read the superblock at?
	 *
	 * NOTE: Need to write an atoi equivalent for off_t - it'll do
	 * for now but only because the superblock is usually not
	 * very far into the partition
	*/
	psDebugAFS->de_psVolume->av_nSuperBlockPos = DAFS_SUPERBLOCK_POS;

	if(num_tokens(paCmdtokens) == 2) {
		psDebugAFS->de_psVolume->av_nSuperBlockPos = atoi(paCmdtokens[1]);
	}

	afs_read_superblock(psDebugAFS->de_nDev, psDebugAFS->de_psVolume, psDebugAFS->de_psSuperBlock);

    printf("\nSuperBlock Info\n===============\n");
    printf("Partition Name: %s\n", psDebugAFS->de_psSuperBlock->as_zName);
    printf("Magic Number 1: %#lx\n", psDebugAFS->de_psSuperBlock->as_nMagic1);
    printf("Byte Order: %ld\n", psDebugAFS->de_psSuperBlock->as_nByteOrder);
    printf("Block Size: %lu\n", psDebugAFS->de_psSuperBlock->as_nBlockSize);
	printf("Block Shift: %lu\n", psDebugAFS->de_psSuperBlock->as_nBlockShift);
    printf("Number of Blocks: %lld\n", psDebugAFS->de_psSuperBlock->as_nNumBlocks);
    printf("Number of Used Blocks: %lld\n", psDebugAFS->de_psSuperBlock->as_nUsedBlocks);
    printf("Inode Size: %ld\n", psDebugAFS->de_psSuperBlock->as_nInodeSize);
    printf("Magic Number 2: %#lx\n", psDebugAFS->de_psSuperBlock->as_nMagic2);
    printf("Block Per Group: %ld\n", psDebugAFS->de_psSuperBlock->as_nBlockPerGroup);
    printf("Alloc Grp Shift: %ld\n", psDebugAFS->de_psSuperBlock->as_nAllocGrpShift);
    printf("Alloc Group Count: %ld\n", psDebugAFS->de_psSuperBlock->as_nAllocGroupCount);
    printf("Flags: %ld\n", psDebugAFS->de_psSuperBlock->as_nFlags); 

	printf("Log Block group: %ld, start: %u, len: %u\n",
			psDebugAFS->de_psSuperBlock->as_sLogBlock.group,
			psDebugAFS->de_psSuperBlock->as_sLogBlock.start,
			psDebugAFS->de_psSuperBlock->as_sLogBlock.len);

	printf("Log Start: %lld\n", psDebugAFS->de_psSuperBlock->as_nLogStart);
	printf("Valid Log Blocks: %d\n", psDebugAFS->de_psSuperBlock->as_nValidLogBlocks);
	printf("Log Size: %d\n", psDebugAFS->de_psSuperBlock->as_nLogSize);
	printf("Magic Number 3: %#lx\n", psDebugAFS->de_psSuperBlock->as_nMagic3);

	printf("Root Dir group: %ld, start: %u, len: %u\n",
			psDebugAFS->de_psSuperBlock->as_sRootDir.group,
			psDebugAFS->de_psSuperBlock->as_sRootDir.start,
			psDebugAFS->de_psSuperBlock->as_sRootDir.len);
	printf("Deleted Files group: %ld, start: %u, len: %u\n",
			psDebugAFS->de_psSuperBlock->as_sDeletedFiles.group,
			psDebugAFS->de_psSuperBlock->as_sDeletedFiles.start,
			psDebugAFS->de_psSuperBlock->as_sDeletedFiles.len);
	printf("Index Dir group: %ld, start: %u, len: %u\n",
			psDebugAFS->de_psSuperBlock->as_sIndexDir.group,
			psDebugAFS->de_psSuperBlock->as_sIndexDir.start,
			psDebugAFS->de_psSuperBlock->as_sIndexDir.len);

	printf("Boot Loader Size: %d\n", psDebugAFS->de_psSuperBlock->as_nBootLoaderSize);
	printf("Pad length: %d\n", sizeof(psDebugAFS->de_psSuperBlock->as_anPad));
}

void cmd_show_inode_stats(char **paCmdtokens, DebugAFS_s *psDebugAFS)
{
	BlockRun_s	psBlockRun;
	AfsInode_s	*psInode;
	int i, n;

	if(psDebugAFS->de_nDev == DFS_DEVICE_CLOSED) {
		printf("Error: No device currently open, please use open\n");
		return;
	}

	if(num_tokens(paCmdtokens) != 3) {
		cmd_show_help(NULL, HELP_SHOW_INODE_STATS);
		return;
	}

	/* Do we need to read in the superblock?
	 * FIXME: Tries to read in but doesn't test to make sure of success */
	if(psDebugAFS->de_psVolume->av_nSuperBlockPos == DFS_NOREAD_SBLOCK) {
		psDebugAFS->de_psVolume->av_nSuperBlockPos = DAFS_SUPERBLOCK_POS;

		afs_read_superblock(psDebugAFS->de_nDev, psDebugAFS->de_psVolume, psDebugAFS->de_psSuperBlock);
	}

	psInode = malloc(psDebugAFS->de_psSuperBlock->as_nBlockSize);
 
	if(psInode == NULL) {
		printf("Error: cmd_show_inode_stats() no memory for inode\n");
		return;
	}

	/* FINISH ME */
	psBlockRun.group = atoi(paCmdtokens[1]);
	psBlockRun.start = atoi(paCmdtokens[2]);
	psBlockRun.len = 1;

	afs_read_inode(psDebugAFS->de_nDev, psDebugAFS->de_psSuperBlock, &psBlockRun, psInode);

	printf("\nInode Info\n==========\n");
	printf("Magic Number 1: %#lx\n", psInode->ai_nMagic1);
	printf("Inode Num group: %ld, start: %u, len: %u\n",
			psInode->ai_sInodeNum.group,
			psInode->ai_sInodeNum.start,
			psInode->ai_sInodeNum.len);
	printf("UID: %ld\n", psInode->ai_nUID);
	printf("GID: %ld\n", psInode->ai_nGID);

	printf("Mode: (%#lo) ", psInode->ai_nMode);
	for(i = false, n = 0; asInodeModes[n].bi_nFlag != (long int)NULL; n++) {
		if((psInode->ai_nMode & asInodeModes[n].bi_nFlag) == asInodeModes[n].bi_nFlag) {
			printf("%s%s", i == true ? " | " : "", asInodeModes[n].bi_pzName);
			i = true;
		}
	}

	printf("\nFlags: (%#lo) ", psInode->ai_nFlags);
	for(i = false, n = 0; asInodeFlags[n].bi_nFlag != (long int)NULL; n++) {
		if((psInode->ai_nFlags & asInodeFlags[n].bi_nFlag) == asInodeFlags[n].bi_nFlag) {
			printf("%s%s", i == true ? " | " : "", asInodeFlags[n].bi_pzName);
			i = true;
		}
	}

	printf("\nLink Count: %d\n", psInode->ai_nLinkCount);
	printf("Create Time: %lld\n", psInode->ai_nCreateTime);
	printf("Modified Time: %lld\n", psInode->ai_nModifiedTime);
	printf("Inode Parent group: %ld, start: %u, len: %u\n",
			psInode->ai_sParent.group,
			psInode->ai_sParent.start,
			psInode->ai_sParent.len);
	printf("Inode Attribute Dir group: %ld, start: %u, len: %u\n",
			psInode->ai_sAttribDir.group,
			psInode->ai_sAttribDir.start,
			psInode->ai_sAttribDir.len);

	printf("Index Type: ");
	for(n = 0; asInodeIType[n].bi_nFlag != (long int)NULL &&
			psInode->ai_nIndexType != asInodeIType[n].bi_nFlag; n++);
	printf("%s\n", asInodeIType[n].bi_pzName);

	printf("Inode Size: %ld\n", psInode->ai_nInodeSize);

	/* Don't know what a V Node is used for yet */
	if(psInode->ai_psVNode == NULL)
		printf("V Node: NULL\n");
	else
		printf("V Node: %#x\n", (int)psInode->ai_psVNode);

	printf("Data Stream {\n");
	i = sizeof(psInode->ai_sData.ds_asDirect) / sizeof(BlockRun_s);

	for(n = 0; n < i; n++) {
		printf("\tDirect[%d] group: %ld, start: %u, len: %u\n", n,
				psInode->ai_sData.ds_asDirect[n].group,
				psInode->ai_sData.ds_asDirect[n].start,
				psInode->ai_sData.ds_asDirect[n].len);
	}
	printf("\tMax Direct Range: %lld\n\n", psInode->ai_sData.ds_nMaxDirectRange);

	printf("\tIndirect group: %ld, start: %u, len: %u\n",
			psInode->ai_sData.ds_sIndirect.group,
			psInode->ai_sData.ds_sIndirect.start,
			psInode->ai_sData.ds_sIndirect.len);
	printf("\tMax Indirect Range: %lld\n\n", psInode->ai_sData.ds_nMaxIndirectRange);

	printf("\tDouble Indirect group: %ld, start: %u, len: %u\n",
			psInode->ai_sData.ds_sDoubleIndirect.group,
			psInode->ai_sData.ds_sDoubleIndirect.start,
			psInode->ai_sData.ds_sDoubleIndirect.len);
	printf("\tMax Double Indirect Range: %lld\n\n", psInode->ai_sData.ds_nMaxDoubleIndirectRange);

	printf("\tSize: %lld\n", psInode->ai_sData.ds_nSize);
	printf("}\n");

	printf("Pad length: %d\n", sizeof(psInode->ai_anPad));
	printf("Small Data length: %d\n", sizeof(psInode->ai_SmallData));

	free(psInode);
}

void cmd_exit(char **paCmdtokens, DebugAFS_s *psDebugAFS) {

	if(psDebugAFS->de_nDev != DFS_DEVICE_CLOSED)
		cmd_close_device(paCmdtokens, psDebugAFS);

	exit(0);
}


struct CmdOptions_s {
	char *cmd;
	void (*pt)(void **, void *);
} aCmdOptions_s[] = {
	{"super_stats", (void (*)(void **, void *))cmd_show_superblock_stats},
	{"inode_stats", (void (*)(void **, void *))cmd_show_inode_stats},
	{"status", (void (*)(void **, void *))cmd_show_status },
	{"close", (void (*)(void **, void *))cmd_close_device },
	{"open", (void (*)(void **, void *))cmd_open_device },
	{"v", (void (*)(void **, void *))cmd_show_version },
	{"q", (void (*)(void **, void *))cmd_exit },
	{ "h", (void (*)(void **, void *))cmd_show_help },
	{ "?", (void (*)(void **, void *))cmd_show_help },
	{ NULL, NULL }	/* End up here if no match made, could have a fall through function */
};


int main()
{
	int		i;
	char		*pCmdLine, *temp, *paCmdtokens[MAXCMDTOKENS];
	DebugAFS_s	*psDebugAFS;

	pCmdLine = malloc(MAXCMDLENGTH);
	psDebugAFS = (DebugAFS_s *)malloc(sizeof(DebugAFS_s));
	psDebugAFS->de_nDev = DFS_DEVICE_CLOSED;
	psDebugAFS->de_psVolume = (AfsVolume_s *)malloc(sizeof(AfsVolume_s));
	psDebugAFS->de_psVolume->av_nSuperBlockPos = DFS_NOREAD_SBLOCK;
	psDebugAFS->de_psSuperBlock = (AfsSuperBlock_s *)malloc(AFS_SUPERBLOCK_SIZE);

	paCmdtokens[0] = NULL;

	/* Initial warning, only people who know EXACTLY what they're
	 * doing should use debugafs */
	printf("\nWARNING WARNING WARNING WARNING WARNING\n");
	printf("!! Only use this application if you  !!\n");
	printf("!! know EXACTLY what you're doing... !!\n");
	printf("WARNING WARNING WARNING WARNING WARNING\n");

	cmd_show_version(NULL, NULL);
	cmd_show_help(NULL, HELP_STANDARD);

	/* Righto, this code is a little better for handling interactive
	 * mode. Aren't function pointers wonderful? ;) */
	while(1) {
		printf("debugafs: ");
		fgets(pCmdLine, MAXCMDLENGTH, stdin);

		/* Strip out enter */
		pCmdLine[strlen(pCmdLine) - 1] = '\0';

		if(strlen(pCmdLine) == 0)
			continue;

		/* Tokenize the command string using spaces as the separator.
		 * NOTE: Using strtok may not be thread safe, need to look at
		 * strtok_r */
		for(i = 0; paCmdtokens[i] != NULL; i++)
			free(paCmdtokens[i]);

		for(i = 0, temp = strtok(pCmdLine, " "); temp != NULL && i < MAXCMDTOKENS;
		  temp = strtok(NULL, " "), i++)
			paCmdtokens[i] = strdup(temp);

		paCmdtokens[i] = NULL;

		/* Figure out which command and therefore which function to call */
		for(i = 0; aCmdOptions_s[i].cmd != NULL &&
				strcasecmp(aCmdOptions_s[i].cmd, paCmdtokens[0]); i++);

		if(aCmdOptions_s[i].cmd == NULL)
			cmd_show_help(NULL, HELP_UNKNOWN);
		else
			aCmdOptions_s[i].pt((void **)paCmdtokens, psDebugAFS);
	}

	return 0;
}
